/* ====================================================================================================================

  The copyright in this software is being made available under the License included below.
  This software may be subject to other third party and   contributor rights, including patent rights, and no such
  rights are granted under this license.

  Copyright (c) 2010, SAMSUNG ELECTRONICS CO., LTD. and BRITISH BROADCASTING CORPORATION
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of SAMSUNG ELECTRONICS CO., LTD. nor the name of the BRITISH BROADCASTING CORPORATION
      may be used to endorse or promote products derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

/** \file     TComPic.cpp
    \brief    picture class
*/

#include "TComPic.h"

// ====================================================================================================================
// Constructor / destructor / create / destroy
// ====================================================================================================================

TComPic::TComPic()
{
  m_apcPicSym         = NULL;
  m_apcPicYuv[0]      = NULL;
  m_apcPicYuv[1]      = NULL;
#if HHI_INTERP_FILTER
  m_apcPicYuv[2]      = NULL;
#endif
#ifdef _ENABLE_JCTVC_B031
#if HHI_INTERP_FILTER
  m_apcPicYuv[3]      = NULL;
#else
  m_apcPicYuv[2]      = NULL;
#endif

#endif
  m_pcPicYuvPred      = NULL;
  m_pcPicYuvResi      = NULL;

  m_bReconstructed    = false;
}

TComPic::~TComPic()
{
}

Void TComPic::create( Int iWidth, Int iHeight, UInt uiMaxWidth, UInt uiMaxHeight, UInt uiMaxDepth, Bool bIsVirtual )
{
  m_apcPicSym     = new TComPicSym;  m_apcPicSym   ->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth );
  if (!bIsVirtual)
  {
    m_apcPicYuv[0]  = new TComPicYuv;  m_apcPicYuv[0]->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth );
  }
  m_apcPicYuv[1]  = new TComPicYuv;  m_apcPicYuv[1]->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth );

#if HHI_INTERP_FILTER
  // buffer for filtered reconstructed pic
  m_apcPicYuv[2]  = new TComPicYuv;  m_apcPicYuv[2]->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth );
#endif
#ifdef _ENABLE_JCTVC_B031
#if HHI_INTERP_FILTER
  m_apcPicYuv[3]  = new TComPicYuv;  m_apcPicYuv[3]->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth );
#else
  m_apcPicYuv[2]  = new TComPicYuv;  m_apcPicYuv[2]->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth );
#endif
#endif
  return;
}

Void TComPic::destroy()
{
  if (m_apcPicSym)
  {
    m_apcPicSym->destroy();
    delete m_apcPicSym;
    m_apcPicSym = NULL;
  }

  if (m_apcPicYuv[0])
  {
    m_apcPicYuv[0]->destroy();
    delete m_apcPicYuv[0];
    m_apcPicYuv[0]  = NULL;
  }

  if (m_apcPicYuv[1])
  {
    m_apcPicYuv[1]->destroy();
    delete m_apcPicYuv[1];
    m_apcPicYuv[1]  = NULL;
  }

#if HHI_INTERP_FILTER
  if (m_apcPicYuv[2])
  {
    m_apcPicYuv[2]->destroy();
    delete m_apcPicYuv[2];
    m_apcPicYuv[2]  = NULL;
  }
#endif
#ifdef _ENABLE_JCTVC_B031
#if HHI_INTERP_FILTER
  if (m_apcPicYuv[3])
  {
    m_apcPicYuv[3]->destroy();
    delete m_apcPicYuv[3];
    m_apcPicYuv[3]  = NULL;
  }
#else
  if (m_apcPicYuv[2])
  {
    m_apcPicYuv[2]->destroy();
    delete m_apcPicYuv[2];
    m_apcPicYuv[2]  = NULL;
  }
#endif
#endif

}

#ifdef _ENABLE_JCTVC_B031
Void TComPic::createRestrictedPicture()
{
  TComPicYuv* pcPicYuvRec        = this->getPicYuvRec();
  TComPicYuv* pcPicYuvRestricted = this->getPicYuvRestricted();

  UInt uiBoundaryPosition = this->getBoundaryPosition();
  
  // Luma
  Pel *pSrc = pcPicYuvRec->getBufY();
  Pel *pDst = pcPicYuvRestricted->getBufY();

  for(Int y = 0; y < (pcPicYuvRec->getHeight() + ((pcPicYuvRec->getLumaMarginY()) << 1)); y ++)
  {
    Pel pTemp = *(pSrc + pcPicYuvRec->getLumaMargin() + uiBoundaryPosition - 1);
    for(Int x = 0; x < (pcPicYuvRec->getWidth() + ((pcPicYuvRec->getLumaMargin()) << 1)); x ++)
    {
      if(x < (uiBoundaryPosition + pcPicYuvRec->getLumaMargin()))
      {
        pDst[x] = pSrc[x];
      }
      else
      {
        pDst[x] = pTemp;
      }
    }
    pSrc += pcPicYuvRec->getStride();
    pDst += pcPicYuvRec->getStride();
  }

  uiBoundaryPosition = (uiBoundaryPosition >> 1);

  // U
  pSrc = pcPicYuvRec->getBufU();
  pDst = pcPicYuvRestricted->getBufU();

  for(Int y = 0; y < ((pcPicYuvRec->getHeight() >> 1) + ((pcPicYuvRec->getChromaMarginY()) << 1)); y ++)
  {
    Pel pTemp = *(pSrc + pcPicYuvRec->getChromaMargin() + uiBoundaryPosition - 1);
    for(Int x = 0; x < ((pcPicYuvRec->getWidth() >> 1) + ((pcPicYuvRec->getChromaMargin()) << 1)); x ++)
    {
      if(x < (uiBoundaryPosition + pcPicYuvRec->getChromaMargin()))
      {
        pDst[x] = pSrc[x];
      }
      else
      {
        pDst[x] = pTemp;
      }
    }
    pSrc += pcPicYuvRec->getCStride();
    pDst += pcPicYuvRec->getCStride();
  }

  // V
  pSrc = pcPicYuvRec->getBufV();
  pDst = pcPicYuvRestricted->getBufV();

  for(Int y = 0; y < ((pcPicYuvRec->getHeight() >> 1) + ((pcPicYuvRec->getChromaMarginY()) << 1)); y ++)
  {
    Pel pTemp = *(pSrc + pcPicYuvRec->getChromaMargin() + uiBoundaryPosition - 1);
    for(Int x = 0; x < ((pcPicYuvRec->getWidth() >> 1) + ((pcPicYuvRec->getChromaMargin()) << 1)); x ++)
    {
      if(x < (uiBoundaryPosition + pcPicYuvRec->getChromaMargin()))
      {
        pDst[x] = pSrc[x];
      }
      else
      {
        pDst[x] = pTemp;
      }
    }
    pSrc += pcPicYuvRec->getCStride();
    pDst += pcPicYuvRec->getCStride();
  }
}
#endif
