/* ====================================================================================================================

  The copyright in this software is being made available under the License included below.
  This software may be subject to other third party and   contributor rights, including patent rights, and no such
  rights are granted under this license.

  Copyright (c) 2010, SAMSUNG ELECTRONICS CO., LTD. and BRITISH BROADCASTING CORPORATION
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification, are permitted only for
  the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
  promoting such standards. The following conditions are required to be met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and
      the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
      the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of SAMSUNG ELECTRONICS CO., LTD. nor the name of the BRITISH BROADCASTING CORPORATION
      may be used to endorse or promote products derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

/** \file     TComPic.h
    \brief    picture class (header)
*/

#ifndef __TCOMPIC__
#define __TCOMPIC__

// Include files
#include "CommonDef.h"
#include "TComPicSym.h"
#include "TComPicYuv.h"
#include "TComBitStream.h"

// ====================================================================================================================
// Class definition
// ====================================================================================================================

/// picture class (symbol + YUV buffers)
class TComPic
{
private:
  TComPicSym*           m_apcPicSym;              //  Symbol

#if HHI_INTERP_FILTER
#ifdef _ENABLE_JCTVC_B031
  TComPicYuv*           m_apcPicYuv[4];           //  Texture,  0:org / 1:rec / 2:rec-filtered
#else
  TComPicYuv*           m_apcPicYuv[3];           //  Texture,  0:org / 1:rec / 2:rec-filtered
#endif
#else
#ifdef _ENABLE_JCTVC_B031
  TComPicYuv*           m_apcPicYuv[3];           //  Texture,  0:org / 1:rec
#else
  TComPicYuv*           m_apcPicYuv[2];           //  Texture,  0:org / 1:rec
#endif
#endif

  TComPicYuv*           m_pcPicYuvPred;           //  Prediction
  TComPicYuv*           m_pcPicYuvResi;           //  Residual
  Bool                  m_bReconstructed;

#ifdef _MULTI_SLICE
  UInt                  m_uiCurrSliceIdx;         // Index of current slice
  UInt                  m_uiLastLctbIdx;          // Index of last Lctb in current slice
#endif

#ifdef _INTRA_REFRESH
  UInt                  m_uiBoundaryPosition;
  UInt                  m_uiRegionIdx;
#endif
public:
  TComPic();
  virtual ~TComPic();

  Void          create( Int iWidth, Int iHeight, UInt uiMaxWidth, UInt uiMaxHeight, UInt uiMaxDepth, Bool bIsVirtual = false );
  Void          destroy();

  TComPicSym*   getPicSym()           { return  m_apcPicSym;    }
#ifdef _MULTI_SLICE
  TComSlice*    getSlice(Int i)       { return  m_apcPicSym->getSlice(i);  }
  Int           getPOC()              { return  m_apcPicSym->getSlice(m_uiCurrSliceIdx)->getPOC();  }
  Bool          getDRBFlag()          { return  m_apcPicSym->getSlice(m_uiCurrSliceIdx)->getDRBFlag();  }
  Int           getERBIndex()         { return  m_apcPicSym->getSlice(m_uiCurrSliceIdx)->getERBIndex();  }
#else
  TComSlice*    getSlice()            { return  m_apcPicSym->getSlice();  }
  Int           getPOC()              { return  m_apcPicSym->getSlice()->getPOC();  }
  Bool          getDRBFlag()          { return  m_apcPicSym->getSlice()->getDRBFlag();  }
  Int           getERBIndex()         { return  m_apcPicSym->getSlice()->getERBIndex();  }
#endif
  TComDataCU*&  getCU( UInt uiCUAddr )  { return  m_apcPicSym->getCU( uiCUAddr ); }

  TComPicYuv*   getPicYuvOrg()        { return  m_apcPicYuv[0]; }
  TComPicYuv*   getPicYuvRec()        { return  m_apcPicYuv[1]; }

#if HHI_INTERP_FILTER
  TComPicYuv*   getPicYuvRecFilt()    { return  m_apcPicYuv[2]; }
#endif
#ifdef _ENABLE_JCTVC_B031
#if HHI_INTERP_FILTER
  TComPicYuv*   getPicYuvRestricted()    { return  m_apcPicYuv[3]; }
#else
  TComPicYuv*   getPicYuvRestricted()    { return  m_apcPicYuv[2]; }
#endif
#endif
#ifdef _ENABLE_JCTVC_B031
  Void          createRestrictedPicture();
#endif

  TComPicYuv*   getPicYuvPred()       { return  m_pcPicYuvPred; }
  TComPicYuv*   getPicYuvResi()       { return  m_pcPicYuvResi; }
  Void          setPicYuvPred( TComPicYuv* pcPicYuv )       { m_pcPicYuvPred = pcPicYuv; }
  Void          setPicYuvResi( TComPicYuv* pcPicYuv )       { m_pcPicYuvResi = pcPicYuv; }

  UInt          getNumCUsInFrame()      { return m_apcPicSym->getNumberOfCUsInFrame(); }
  UInt          getNumPartInWidth()     { return m_apcPicSym->getNumPartInWidth();     }
  UInt          getNumPartInHeight()    { return m_apcPicSym->getNumPartInHeight();    }
  UInt          getNumPartInCU()        { return m_apcPicSym->getNumPartition();       }
  UInt          getFrameWidthInCU()     { return m_apcPicSym->getFrameWidthInCU();     }
  UInt          getFrameHeightInCU()    { return m_apcPicSym->getFrameHeightInCU();    }
  UInt          getMinCUWidth()         { return m_apcPicSym->getMinCUWidth();         }
  UInt          getMinCUHeight()        { return m_apcPicSym->getMinCUHeight();        }

  UInt          getParPelX(UChar uhPartIdx) { return getParPelX(uhPartIdx); }
  UInt          getParPelY(UChar uhPartIdx) { return getParPelX(uhPartIdx); }

  Int           getStride()           { return m_apcPicYuv[1]->getStride(); }
  Int           getCStride()          { return m_apcPicYuv[1]->getCStride(); }

  Void          setReconMark (Bool b) { m_bReconstructed = b;     }
  Bool          getReconMark ()       { return m_bReconstructed;  }

#ifdef _MULTI_SLICE
  UInt          getCurrSliceIdx()            { return m_uiCurrSliceIdx;                }
  UInt          getLastLctbIdx()             { return m_uiLastLctbIdx;                 }
  Void          setCurrSliceIdx(UInt i)      { m_uiCurrSliceIdx = i;                   }
  Void          setLastLctbIdx(UInt i)       { m_uiLastLctbIdx = i;                    }
#endif
#ifdef _INTRA_REFRESH
  UInt          getBoundaryPosition()        { return m_uiBoundaryPosition;            }
  Void          setBoundaryPosition(UInt i)  { m_uiBoundaryPosition = i;               }
  UInt          getRegionIdx()               { return m_uiRegionIdx;                   }
  Void          setRegionIdx(UInt i)         { m_uiRegionIdx = i;                      }
#endif
};// END CLASS DEFINITION TComPic


#endif // __TCOMPIC__

